/** C API for cres
 *
 * See `examples/cres.c` and `examples/user_distance.c` for for usage.
 * The main function is `cres_run`.
 *
 * Functions return an integer, with `0` indicating success and
 * everything else indicating an error. Errors can be accessed with
 * `cres_get_last_err` and `cres_print_last_err`.
 *
 * License: GPL 3.0 or later
 * Author: Andreas Maier <andreas.martin.maier@desy.de>
*/

#ifndef CRES_H
#define CRES_H

#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>

/**
 * Jet clustering algorithms
 */
typedef enum JetAlgorithm {
  /**
   * The [anti-kt](https://arxiv.org/abs/0802.1189) algorithm
   */
  AntiKt,
  /**
   * The [Cambridge](https://arxiv.org/abs/hep-ph/9707323)/[Aachen](https://arxiv.org/abs/hep-ph/9907280) algorithm
   */
  CambridgeAachen,
  /**
   * The [kt](https://arxiv.org/abs/hep-ph/9305266) algorithm
   */
  Kt,
} JetAlgorithm;

/**
 * Nearest-neighbour search algorithms
 */
typedef enum Search {
  /**
   * Vantage point tree search
   */
  Tree,
  /**
   * Naive search
   */
  Naive,
} Search;

typedef struct NoObserver NoObserver;

/**
 * Disable unweighting
 */
typedef struct NoUnweighter NoUnweighter;

/**
 * Four-momentum [E, px, py, pz]
 */
typedef double FourMomentum[4];

/**
 * View into a particle set
 *
 * Changing any member does not change the original particle set
 */
typedef struct TypeSetView {
  /**
   * Particle id
   *
   * This is usually the id according to the
   * [PDG Monte Carlo Particle Numbering Scheme](https://pdg.lbl.gov/2021/mcdata/mc_particle_id_contents.html)
   */
  int32_t pid;
  /**
   * Four-momenta of the particles
   */
  const FourMomentum *momenta;
  /**
   * Number of particles
   */
  uintptr_t n_momenta;
} TypeSetView;

/**
 * View into an event
 *
 * Changing any of the members does not change the original event.
 */
typedef struct EventView {
  /**
   * Event id
   */
  uintptr_t id;
  /**
   * Event weight
   */
  double weight;
  /**
   * Sets of particles of a given type
   */
  const struct TypeSetView *type_sets;
  /**
   * Number of particle sets
   */
  uintptr_t n_type_sets;
} EventView;

/**
 * User-defined distance function
 */
typedef struct DistanceFn {
  /**
   * The distance function
   *
   * This has to be a *thread-safe* function that _may never return NaN_.
   * The first argument is a pointer to the `data` member of this struct.
   * The remaining arguments are the events for which we compute the distance.
   */
  double (*fun)(void*, const struct EventView*, const struct EventView*);
  /**
   * Arbitrary data used by the distance function
   */
  void *data;
} DistanceFn;

typedef struct JetDefinition {
  /**
   * Jet algorithm
   */
  enum JetAlgorithm algorithm;
  /**
   * Jet radius parameter
   */
  double radius;
  /**
   * Minimum jet transverse momentum
   */
  double min_pt;
} JetDefinition;

/**
 * Resampling options
 */
typedef struct Opt {
  /**
   * Names of input files
   *
   * Input files should be in HepMC2 format,
   * possibly compressed with bzip2, gzip, lz4, or zstd
   */
  char **infiles;
  /**
   * Number of input files
   */
  uintptr_t n_infiles;
  /**
   * Name of HepMC output file
   */
  char *outfile;
  /**
   * Which distance function to use
   *
   * If set to `NULL`, the default distance function from
   * [arXiv:2109.07851](https://arxiv.org/abs/2109.07851)
   * is used
   */
  struct DistanceFn *distance;
  /**
   * Extra contribution to distance proportional to difference in pt
   *
   * This parameter is ignored when using a custom distance. Otherwise,
   * it corresponds to the τ parameter of
   * [arXiv:2109.07851](https://arxiv.org/abs/2109.07851)
   */
  double ptweight;
  /**
   * Jet definition
   */
  struct JetDefinition jet_def;
  /**
   * Algorithm for finding nearest-neigbour events,
   */
  enum Search neighbour_search;
  /**
   * Number of partitions
   */
  uint32_t num_partitions;
  /**
   * Maximum cell radius
   *
   * Set to INFINITY for unlimited cell sizes
   */
  double max_cell_size;
} Opt;





#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

/**
 * Run the cell resampler with the given options
 *
 * # Return values
 *
 * - `0`: success
 * - Non-zero: an error occurred, check with `cres_get_last_err` or
 *   `cres_print_last_err`
 */
__attribute__((warn_unused_result)) int32_t cres_run(const struct Opt *opt);

/**
 * Print the last `cres` error that occurred to the standard error output.
 *
 * Calling this function repeatedly without any intermediate errors
 * will always give the same output. If no errors have occurred so far,
 * this function will do nothing.
 */
void cres_print_last_err(void);

/**
 * Copy the last `cres` error message to the provided buffer.
 *
 * # Return values
 * -  `0`: success
 * - `-1`: rust panic
 * - any other non-zero: The provided buffer was too small and the
 *   error message was truncated. The return value indicates the
 *   required buffer size.
 */
__attribute__((warn_unused_result)) int32_t cres_get_last_err(char *buf, uintptr_t buflen);

/**
 * Initialise logger from environment variable
 *
 * The name of the environment variable must be a valid utf8 string.
 *
 * # Safety
 *
 * The passed pointer must point to a valid null-terminated C string.
 *
 * # Return values
 *
 * - `0`: success
 * - `-1`: rust panic, check with `cres_get_last_err` or `cres_print_last_err`
 */
__attribute__((warn_unused_result)) int32_t cres_logger_from_env(const char *env_var);

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus

#endif /* CRES_H */
